/*
 * Screen and keyboard handling for curses.
 * */

#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <ncurses/ncurses.h>

#include "tdisplay.h"
#include "include/keycode.h"

#include <sys/ioctl.h>
#include <signal.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>



static int attr_map[256];
int screen_init(void);

/* Keyboard handling functions */

int key_init(void)
{
	return 1;
}

void key_free(void)
{
}

static int local_key = 0;

static int get_key()
{
	int k = getch();

	if (k == ERR)
		return 0;
	
#ifdef SIGTSTP1
	if (k == KEY_SUSPEND)
		raise(SIGTSTP);
#endif /* SIGTSTP */
	
	switch (k) {
	case KEY_LEFT:
		k = K_LEFT;
		break;
	case KEY_RIGHT:
		k = K_RIGHT;
		break;
	case KEY_UP:
		k = K_UP;
		break;
	case KEY_DOWN:
		k = K_DOWN;
		break;
	case KEY_HOME:
		k = K_HOME;
		break;
	case KEY_END:
		k = K_END;
		break;
	case KEY_IC:
		k = K_INS;
		break;
	case KEY_DC:
		k = K_DEL;
		break;
	case KEY_NPAGE:
		k = K_PGDN;
		break;
	case KEY_PPAGE:
		k = K_PGUP;
		break;
//      case KEY_B2: k = key_5; break;
//      case KEY_MOUSE: k = key_mouse; break;
	case 127:
	case KEY_BACKSPACE:
		k = '\b';
		break;
	case '\r':
	case '\n':
		return '\r';
	default:
		if (k >= KEY_F(0) && k <= KEY_F(30))
			k = k - KEY_F(0) + 330;
		break;
	}

	if (k == 27) {
		k = getch();
		if (k != ERR && k != 27) {
			if ((k | ' ') >= 'a' && (k | ' ') <= 'z')
				return K_ALT_A + ((k | ' ') - 'a');
			else
				return k;
		} else
			return 27;
	} else
		return k;
}

int key_ready(void)
{
	if (local_key)
		return local_key;
	local_key = get_key();
	if (local_key)
		return local_key;
	return 0;
}

int key_get(void)
{
	int k = local_key;
	local_key = 0;
	return k;
}


/* Screen handling functions */
/* If you want to port it to another system,
   you must provide few simple conio.h functions.
   (gotoxy(),putch(),_setcursortype(),textattr())
*/

// schamelessly stolen from linux tvision port
static void init_color_colors()
{
	int vgac[] = {
		COLOR_BLACK, COLOR_BLUE, COLOR_GREEN, COLOR_CYAN,
		COLOR_RED, COLOR_MAGENTA, COLOR_YELLOW, COLOR_WHITE,
	};
	int back, fore, bold, i;

	start_color();

	i = 0;
	for (back = 0; back < 8; back++)
		for (fore = 7; fore >= 0; fore--)
			init_pair(i++, vgac[fore], vgac[back]);

	for (i = 0; i < 256; i++) {
		back = (i >> 4) & 0x07;
		bold = i & 0x08;
		fore = i & 0x07;
		attr_map[i] = COLOR_PAIR((7 - fore) + back * 8);
		if (bold)
			attr_map[i] |= A_BOLD;
	}
}

static void init_mono_colors()
{
	int i;
	// subjective brightness of vga colors to decide whatever
	// to use A_REVERSE and/or A_BOLD
	int vgab[] = {
		 /*BLACK*/ 0, /*BLUE*/ 4, /*GREEN*/ 5, /*CYAN*/ 8,
		 /*RED*/ 3,
		 /*MAGENTA*/ 7, /*BROWN*/ 2, /* LIGHTGRAY */ 11,
		 /*DARKGRAY*/ 1,
		 /*LIGHTBLUE*/ 6, /*LIGHTGREEN*/ 12, /*LIGHTCYAN*/ 13,
		 /*LIGHTRED*/ 9, /*LIGHTMAGENTA*/ 10, /*YELLOW*/ 14,
		 /*WHITE*/ 15
	};
	for (i = 0; i < 256; i++) {
		attr_map[i] =
		    vgab[(i >> 4) & 0xf] > vgab[i & 0xf] ? A_REVERSE : 0;
		if ((i & 0xf) > 11)
			attr_map[i] |= A_BOLD;
	}
}

static void my_init_color()
{
	if (has_colors())
		init_color_colors();
	else
		init_mono_colors();
}

void screen_free(void)
{
	endwin();
	printf("\r");

	return;
}

void setenvs(int h, int w)
{
	char s[4];
	sprintf(s,"%d", h);
	setenv("LINES", s, 1);
		
	sprintf(s,"%d", w);
	setenv("COLUMNS", s, 1);
}

#define MY_STYLE

#ifdef MY_STYLE
static void resize_handler(int sig)
{
	struct winsize size;
	
	(void) sig;
	
	if (ioctl(1, TIOCGWINSZ, &size) == 0) {
		screen_free();
		setenvs(size.ws_row, size.ws_col);
		screen_init();
		t_resize();
	}
}
#endif

#ifdef MC_STYLE
static void resize_handler(int sig)
{
	struct winsize winsz;

	winsz.ws_col = winsz.ws_row = 0;
	ioctl (0, TIOCGWINSZ, &winsz);
	if (winsz.ws_col && winsz.ws_row){
		resizeterm(winsz.ws_row, winsz.ws_col);
		clearok(stdscr,TRUE);	
		COLS = winsz.ws_col;
		LINES = winsz.ws_row;
		t_resize();
	}
}
#endif

#ifndef MC_STYLE 
#ifndef MY_STYLE
static void resize_handler(int sig)
{
	struct winsize w;
	char *cp;
	int fd;
	
	LINES = COLS = -1;

	(void) sig;

	if ((fd = open ("/dev/tty", O_RDONLY)) != -1) {
		if (ioctl (fd, TIOCGWINSZ, &w) != -1) {
			LINES = w.ws_row;
			COLS = w.ws_col;
		}
		close (fd);
	}
	if (LINES <= 0) {
		if ((cp = getenv ("LINES")) != NULL) 
			LINES = atoi (cp);
		else
			LINES = 24;
	}
	if (COLS <= 0) {
		if ((cp = getenv ("COLS")) != NULL)
			COLS = atoi (cp);
		else
			COLS = 80;
	}


	setenvs(LINES, COLS);
  	resizeterm (LINES, COLS);
	clearok(stdscr, TRUE);	
	t_resize();
}
#endif
#endif


int screen_init(void)
{

#ifndef DOUBLE_ESC
	setenv("ESCDELAY", "100", 0);
#endif
	initscr();
	keypad(stdscr, TRUE);
	raw();
	noecho();
	timeout(0);
	refresh();
	meta(stdscr, TRUE);
	idlok(stdscr, TRUE);
	//mousemask(REPORT_MOUSE_POSITION, 0);

	my_init_color();
	signal(SIGWINCH, resize_handler);

	return 1;
}


int screen_getcols(void)
{
	return COLS;
}

int screen_getrows(void)
{
	return LINES;
}


void screen_gotoxy(int x, int y)
{
	x--;
	y--;

	if (y < 0)
		y = 0;
	if (x < 0)
		x = 0;
	if (x >= COLS)
		x = COLS - 1;
	if (y >= LINES)
		y = LINES - 1;

	move(y, x);
}

#define C_B		1
#define C_UL		2
#define C_H		3
#define C_UR		4
#define C_V		5
#define C_HATCH		6
#define C_LL		7
#define C_LR		8

static chtype map(char c)
{
	switch (c) {
	case C_B:
		return ACS_BOARD;
	case C_UL:
		return ACS_ULCORNER;
	case C_H:
		return ACS_HLINE;
	case C_UR:
		return ACS_URCORNER;
	case C_V:
		return ACS_VLINE;
	case C_HATCH:
		return ACS_CKBOARD;
	case C_LL:
		return ACS_LLCORNER;
	case C_LR:
		return ACS_LRCORNER;
	default:
		return (unsigned char) c;
	}
}
/* curses should know that screen expolode ! ;) */
void screen_force_refresh()
{
	clearok(stdscr, TRUE);
	keypad(stdscr, TRUE);
}

void screen_put2scr(int width, int height, void *buf)
{
	struct point {
		unsigned char c;
		unsigned char attr;
	};
	int x, y;
	struct point *p;

	p = buf;
	for (y = 0; y < height; y++) {
		screen_gotoxy(1, y + 1);
		for (x = 0; x < width; x++) {
			addch(map(p->c) | attr_map[p->attr]);
			p++;
		}
	}
	refresh();
}
void screen_refresh()
{
	refresh();
}
static int local_attr;

void screen_textattr(int attr)
{
	local_attr = attr;
}

void screen_putch(unsigned c)
{
	addch(map(c) | attr_map[local_attr]);
}

/* Returns char representing choosen frame:
   1 2 3     + = +
   4   6 ex. |   |
   7 8 9     + - +
*/


char screen_frame_table[][9] = {
	/* ASCII */
	{'+', '=', '+',
	 '|', ' ', '|',
	 '+', '-', '+'},
	/* SINGLE */
	{C_UL, C_H, C_UR,
	 C_V, C_HATCH, C_V,
	 C_LL, C_H, C_LR},
	/* DOUBLE */
	{C_UL, C_B, C_UR,
	 C_V, C_HATCH, C_V,
	 C_LL, C_H, C_LR}
};

char screen_frame(int frame, int type)
{
	return screen_frame_table[type - 1][frame - 1];
}

void screen_setcursortype(int type)
{
	if (type == TSCR_NOCURSOR)
		move(0, 0);
}



/* Misc architecture-depended functions */

void mk_delay(void)
{
	usleep(20 * 1000);
}
