/* tsys.c

Main module. Initialization of screen and keyboard.
Event handling (queue, sending, getting...).


*/

#include "tsys.h"
#include "twin.h"
#include <stdio.h>

/* main object - tsystem (extern) */
struct tsys *tsystem = NULL;



/* queue of events */
/* It is used for queueing events:
  o from objects to programmer (user space). Like EVENT_TBUTTON_PRESSED
     or something like that.
  o from keyboard to objects

  and also (but I think direct tobj_domethod will be better in most cases)
  o from object to another object (like below)
  o from programmer to objects (if you want to simulate
     some event like key pressed or something like that)
*/

static struct tevent *events = NULL;
static struct tevent *event_last = NULL;



/* Allocate memory and add event to queue */
void tevent_send(struct tobj *sender, struct tobj *address, long data)
{
	struct tevent *event;

	if ((event = malloc(sizeof(struct tevent)))) {
		event->sender = sender;
		event->address = address;
		event->data = data;
		event->next = NULL;

		if (event_last) 
			event_last->next = event;
		
		event_last = event;

		if (events == NULL)
			events = event;
	} else
		tsys_error_exit("ERROR: Couldn't allocate memory (tevent).\n");
}

/* Removing event from queue (YOU must free it later) */
struct tevent *tevent_get(void)
{
	struct tevent *event;

	if (events) {
		event = events;

		events = event->next;
		if (!events)
			event_last = NULL;

		event->next = NULL;
		return event;
	} else
		return NULL;

}

/* Checking if queue isn't empty.
   If isn't, it returns address of first event in queue. */
struct tobj *tevent_check(void)
{
	if (events)
		return events->address;
	else
		return 0;

}
/* gets address of first event in queue, DON'T remove it from list */
struct tevent *tevent_first(void)
{
	if (events)
		return events;
	else
		return NULL;

}
void tevent_free(struct tevent *event)
{
	if (event)
		free(event);
}


/* Allocate memory for tsys struct and initializate keybord and screen.
   Makes root window and return ptr to it. */
struct tobj *tsys_init(void)
{
	if (!(tsystem = calloc(1, sizeof(struct tsys))))
		return 0;

	/* screen init (buffers & archdep) */
	if (!t_init())
		return 0;

	/* root window init */
	tsystem->rootwin = twin_init(NULL, 1, 1,
				     t_screencols(), t_screenrows(), NULL);

	if (!tsystem->rootwin)
		return 0;
	tmset(tsystem->rootwin, TWIN_FRAME, TWIN_FRAME_NONE,
	      TWIN_MOVEABLE, 0L, TOBJ_BGCOLOR, BLACK, NULL);

	tsystem->refresh = 1;
	tsystem->lock = 0;

	/* Keybourd init */
	if (!key_init())
		return 0;

	return tsystem->rootwin;
}

void tsys_garbage(void)
{
	/* freeing root window and all its childs */
	if (tsystem->rootwin)
		tobj_garbage(tsystem->rootwin);
	
	tsystem->garbage=0;
}
/* Freeing all resources */
void tsys_free(void)
{
	/* freeing root window and all its childs */
	if (tsystem->rootwin)
		tobj_free(tsystem->rootwin);

	
	if (events) {
		struct tevent *e;
		while (events) {
			e = tevent_get();
			if (e)
				tevent_free(e);
		}

	}

	tsys_garbage();
	if (tsystem)
		free(tsystem);

	key_free();
	t_free();
}

int tsys_domethod(long method, long data)
{
	(void) data;
	
	switch (method) {
	case TSYS_FORCE_REFRESH :
		tsystem->force_refresh = 1;
		tsystem->refresh = 1;
		break;
	case TSYS_LOCK_SCREEN:
		tsystem->lock = 1;
		t_clrscr();
		t_gotoxy(1,1);
		t_cputs((char *)data);
		t_sync();
		break;
	case TSYS_UNLOCK_SCREEN:
		tsystem->lock = 0;
		break;
	}
	
	return 0;
}

/* Reacting on event or sending it to activate object */
int tsys_doevent(int keycode)
{
	if (tsystem->rootwin && (tobj_doevent(tsystem->rootwin, keycode)))
		return 1;
	else {
		/* do something with event */
		switch (keycode) {

		case K_CTL_C:
			tevent_send(NULL, EVENT_TYPE_USER, EVENT_QUIT);
			return 1;
		}

		/* or return 0, if keycode wasn't used */
		return 0;
	}

}


/* Redrawing whole screen */
void tsys_dorefresh(void)
{

	/* redraw background */

	/* redrawing rootwin and its childs */
	if (tsystem->rootwin) {
		/* turning off cursor, if some object will need it,
		   it will turn it back on. This object should be active,
		   but that isn't checked */
		t_cursortype(TSCR_NOCURSOR);
		tobj_dorefresh(tsystem->rootwin);
	}

	/* if curses should know that whole screen 
	 * must be redraw t_force_refresh() is calling */
	if (tsystem->force_refresh) {
		tsystem->force_refresh = 0;
		t_force_refresh();
	}

	t_sync();

	tsystem->refresh = 0;
	tsystem->sync = 0;
}

/* Main function. If pass some object in args, then keys will be sent
 * only to this object.
  Returns:
  -  new event to user (which must be tevent_free())
  -  NULL - non events to user
*/
struct tevent *tsys_process( struct tobj *obj )
{
	int key ;
	struct tobj *address;
	struct tevent *ret = NULL;

	if (key_ready()) {
		key = key_get();
		/* sending keycode to system,
		   if not used sending to user as event */
		if ((!obj && !tsys_doevent(key)) || tsystem->lock) 
			tevent_send(NULL, EVENT_TYPE_USER, key);
		else if (obj && !tobj_doevent(obj , key))
			tevent_send(NULL, EVENT_TYPE_USER, key);
	}
	/* checking event queue */
	else if ((address = tevent_check())) {
		ret = tevent_get();
		
		/* If event is to user (programmer) */
		if (address == EVENT_TYPE_USER){
			if (ret->sender)
				notify_process( ret->sender , ret->data );
		}
		else {
			/* tsys is a something like postman:
			   it sends events to objects (or to itself) */

			if (address == EVENT_TYPE_INTERNAL)
				tsys_doevent(ret->data);
			else
				tobj_doevent(ret->address, ret->data);

			tevent_free(ret);
			ret = NULL;
		}
	} else if (!tsystem->refresh && !tsystem->sync)
		mk_delay();

	/* some tobj_free(obj) have been done ...
	 * Now we remove all those objects from tree
	 * (we do it only when in main tsys (not mono) loop
	 * and there is no events in queue ) */
	if (!obj && !events && tsystem->garbage)
		tsys_garbage();
		
	if (!tsystem->lock) {
		/* whole screen must be refreshed and synced */
		if (tsystem->refresh)
			tsys_dorefresh();
		/* syncing buffer and screen (small changes on screen) */
		if (tsystem->sync) {
			t_sync_small();
			tsystem->sync = 0;
		}
	}

	return ret;

}
/* this function is usefull when you are performing some operations,
 * but you want to system do refresh */

void tsys_checkrefresh( void )
{
	/* whole screen must be refreshed and synced */
	if (tsystem->refresh)
		tsys_dorefresh();
	/* syncing buffer and screen (small changes on screen) */
	if (tsystem->sync) {
		t_sync_small();
		tsystem->sync = 0;
	}
}

void tsys_error(char *str)
{
	tsys_free();
	fprintf(stderr, str);
}

/* brutal error handling - freeing everything that he can,
   printing error and exit */
void tsys_error_exit(char *str)
{
	tsys_free();
	fprintf(stderr, str);
	exit(10);
}
